"""
Art utilities
"""

# stdlib
from typing import Any, Dict, Optional, Tuple

# external
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

# tdub
from tdub.utils import bin_centers


def setup_style():
    matplotlib.use("Agg")
    matplotlib.rcParams["figure.figsize"] = (6, 5.25)
    matplotlib.rcParams["axes.labelsize"] = 14
    matplotlib.rcParams["font.size"] = 12
    matplotlib.rcParams["xtick.top"] = True
    matplotlib.rcParams["ytick.right"] = True
    matplotlib.rcParams["xtick.direction"] = "in"
    matplotlib.rcParams["ytick.direction"] = "in"
    matplotlib.rcParams["xtick.labelsize"] = 12
    matplotlib.rcParams["ytick.labelsize"] = 12
    matplotlib.rcParams["xtick.minor.visible"] = True
    matplotlib.rcParams["ytick.minor.visible"] = True
    matplotlib.rcParams["xtick.major.width"] = 0.8
    matplotlib.rcParams["xtick.minor.width"] = 0.8
    matplotlib.rcParams["xtick.major.size"] = 7.0
    matplotlib.rcParams["xtick.minor.size"] = 4.0
    matplotlib.rcParams["xtick.major.pad"] = 1.5
    matplotlib.rcParams["xtick.minor.pad"] = 1.4
    matplotlib.rcParams["ytick.major.width"] = 0.8
    matplotlib.rcParams["ytick.minor.width"] = 0.8
    matplotlib.rcParams["ytick.major.size"] = 7.0
    matplotlib.rcParams["ytick.minor.size"] = 4.0
    matplotlib.rcParams["ytick.major.pad"] = 1.5
    matplotlib.rcParams["ytick.minor.pad"] = 1.4
    matplotlib.rcParams["legend.frameon"] = False
    matplotlib.rcParams["legend.numpoints"] = 1
    matplotlib.rcParams["legend.fontsize"] = 11
    matplotlib.rcParams["legend.handlelength"] = 1.5
    matplotlib.rcParams["axes.formatter.limits"] = [-4, 4]
    matplotlib.rcParams["axes.formatter.use_mathtext"] = True


def canvas_from_counts(
    counts: Dict[str, np.ndarray],
    errors: Dict[str, np.ndarray],
    bin_edges: np.ndarray,
    stack_error_band: Optional[Any] = None,
    ratio_error_band: Optional[Any] = None,
    **subplots_kw,
) -> Tuple[plt.Figure, plt.Axes, plt.Axes]:
    """create a plot canvas given a dictionary of counts and bin edges

    The ``counts`` and ``errors`` dictionaries are expected to have
    the following keys:

    - ``"Data"``
    - ``"tW_DR"``
    - ``"ttbar"``
    - ``"Zjets"``
    - ``"Diboson"``
    - ``"MCNP"``

    Parameters
    ----------
    counts : dict(str, np.ndarray)
        a dictionary pairing samples to bin counts
    errors : dict(str, np.ndarray)
        a dictionray pairing samples to bin count errors
    bin_edges : np.ndarray
        the histogram bin edges
    stack_error_band : Any, optional
        todo
    ratio_error_band : Any, optional
        todo
    subplots_kw : dict
        remaining keyword arguments passed to :py:func:`matplotlib.pyplot.subplots`

    Returns
    -------
    fig : :obj:`matplotlib.figure.Figure`
        the matplotlib figure
    ax : :obj:`matplotlib.axes.Axes`
        the matplotlib axes for the histogram stack
    axr : :obj:`matplotlib.axes.Axes`
        the matplotlib axes for the ratio comparison

    """
    centers = bin_centers(bin_edges)
    start, stop = bin_edges[0], bin_edges[-1]
    mc_counts = np.zeros_like(centers, dtype=np.float32)
    mc_errs = np.zeros_like(centers, dtype=np.float32)
    for key in counts.keys():
        if key != "Data":
            mc_counts += counts[key]
            mc_errs += errors[key] ** 2
    mc_errs = np.sqrt(mc_errs)
    ratio = counts["Data"] / mc_counts
    ratio_err = counts["Data"] / (mc_counts ** 2) + np.power(
        counts["Data"] * mc_errs / (mc_counts ** 2), 2
    )
    fig, (ax, axr) = plt.subplots(
        2,
        1,
        sharex=True,
        gridspec_kw=dict(height_ratios=[3.25, 1], hspace=0.025),
        **subplots_kw,
    )
    ax.hist(
        [centers for _ in range(5)],
        bins=bin_edges,
        weights=[
            counts["MCNP"],
            counts["Diboson"],
            counts["Zjets"],
            counts["ttbar"],
            counts["tW_DR"],
        ],
        histtype="stepfilled",
        stacked=True,
        label=["MCNP", "Diboson", "$Z$+jets", "$t\\bar{t}$", "$tW$"],
        color=["#9467bd", "#ff7f0e", "#2ca02c", "#d62728", "#1f77b4"],
    )
    ax.errorbar(
        centers, counts["Data"], yerr=errors["Data"], label="Data", fmt="ko", zorder=999
    )
    axr.plot([start, stop], [1.0, 1.0], color="gray", linestyle="solid", marker=None)
    axr.errorbar(centers, ratio, yerr=ratio_err, fmt="ko", zorder=999)
    axr.set_ylim([0.75, 1.25])
    axr.set_yticks([0.8, 0.9, 1.0, 1.1, 1.2])

    return fig, ax, axr



"""
A module to aid working with histograms
"""

# stdlib
from typing import Tuple, Optional, Union, List, Dict, Any

# ext
import numpy as np
import pandas as pd
from uproot_methods.classes import TH1
from pygram11 import fix1dmw

# tdub
from tdub.utils import categorize_branches


class CustomTH1(TH1.Methods, list):
    """A TH1 like skeleton object"""

    pass


class CustomTAxis:
    """A TAxis like object"""

    def __init__(self, edges: np.ndarray) -> None:
        self._fNbins = len(edges) - 1
        self._fXmin = edges[0]
        self._fXmax = edges[-1]
        self._fXbins = edges.astype(np.float64)


def prepare_padded(
    content: np.ndarray, errors: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Prepare arrays for saving to ROOT histogram with over/underflow

    Paramters
    ---------
    content : :py:obj:`numpy.ndarray`
       the bin contents
    error : :py:obj:`numpy.ndarray`
       the error on the bin content (the square-root of the variances)

    Returns
    -------
    :py:obj:`numpy.ndarray`
       the padded content
    :py:obj:`numpy.ndarray`
       the padded sumw2
    """
    nbins = content.shape[0] + 2
    content_padded = np.empty(nbins, dtype=content.dtype)
    content_padded[1:-1] = content
    content_padded[0] = 0.0
    content_padded[-1] = 0.0
    sumw2_padded = np.empty(nbins, dtype=np.float64)
    sumw2_padded[1:-1] = errors ** 2
    sumw2_padded[0] = 0.0
    sumw2_padded[-1] = 0.0
    return content_padded, sumw2_padded


def arrays2th1(
    content: np.ndarray, error: np.ndarray, bins: np.ndarray, title: str = "none"
) -> CustomTH1:
    """create a TH1-like object built from arrays

    Parameters
    ----------
    content : :py:obj:`numpy.ndarray`
       the bin contents
    error : :py:obj:`numpy.ndarray`
       the error on the bin content (the square-root of the variances)
    bins : :py:obj:`numpy.ndarray`
       the binning definition
    title : str
       title the histogram

    Returns
    -------
    :obj:`CustomTH1`
       the ROOT like histogram object
    """

    output = CustomTH1.__new__(CustomTH1)
    if content.dtype == np.float32:
        output._classname = "TH1F"
    elif content.dtype == np.float64:
        output._classname = "TH1D"
    output._fXaxis = CustomTAxis(bins)
    output._fEntries = content.sum()
    output._fTitle = title

    content_padded, output._fSumw2 = prepare_padded(content, error)
    output.extend(content_padded)

    return output


def df2th1(
    dfc: pd.DataFrame,
    dfe: pd.DataFrame,
    weight_col: Optional[Union[List[str], str]] = None,
) -> Union[CustomTH1, Dict[str, CustomTH1]]:
    """create a TH1-like object built from a dataframe structure

    Parameters
    ----------
    dfc : pandas.DataFrame
       the dataframe holding the bin content
    dfe : pandas.DataFrame
       the dataframe holding the bin errors
    weight_name : str or list(str), optional
       name of the weight(s) (column(s) in the dataframe) to use. If
       ``None``, just ``weight_nominal`` is used. if "ALL", all
       weights are used.

    Returns
    -------
    :obj:`CustomTH1` or dict(str, :obj:`CustomTH1`)
       the ROOT like histogram object(s)

    """
    binning = np.linspace(dfc._xmin, dfc._xmax, dfc._nbins + 1)
    if weight_col is None:
        weight_col = "weight_nominal"
    if isinstance(weight_col, str):
        if weight_col == "ALL":
            res = {}
            for weight_name in dfc.columns:
                res[weight_name] = arrays2th1(
                    dfc[weight_name].to_numpy(),
                    dfe[weight_name].to_numpy(),
                    binning,
                    title=dfc._var_used,
                )
            return res
        else:
            return {
                weight_col: arrays2th1(
                    dfc[weight_col].to_numpy(),
                    dfe[weight_col].to_numpy(),
                    binning,
                    title=dfc._var_used,
                )
            }
    else:
        res = {}
        for weight_name in weight_col:
            res[weight_name] = arrays2th1(
                dfc[weight_name].to_numpy(),
                dfe[weight_name].to_numpy(),
                binning,
                title=dfc._var_used,
            )
        return res


def generate_from_df(
    df: pd.DataFrame,
    var: str,
    bins: int,
    range: Tuple[float, float],
    nominal_weight: bool = True,
    systematic_weights: bool = False,
) -> Any:
    """generate histogram(s) from a dataframe

    Parameters
    ----------
    df : pandas.DataFrame
       the dataframe with our variable and weights of interest
    var : str
       the variable (name of distribution) we want to histogram
    bins : int
       the number of bins
    range : tuple(float, float)
       the axis limits (min, max) for the histogram
    nominal_weight : bool
       histogram the data using the nominal weight
    systematic_weights : bool
       histogram the data using the systematic weights in the dataframe

    Returns
    -------
    tuple(pandas.DataFrame, pandas.DataFrame)
       the resulting histogram bin counts are in the first dataframe
       and the bin uncertainties are in the second frame. the columns
       give the weight used to calculate the histograms


    Examples
    --------

    >>> from tdub.utils import quick_files
    >>> from tdub.frames import raw_dataframe
    >>> from tdub.hist import generate_from_df
    >>> qf = quick_files("/path/to/data")
    >>> df_tW_DR = raw_dataframe(qf["tW_DR"])
    >>> hist_result = generate_from_df(
    ...     df_tW_DR,
    ...     "met",
    ...     bins=20,
    ...     range=(0.0, 200.0),
    ...     systematic_weights=True
    ... )

    """
    weight_cols: List[str] = []
    if nominal_weight:
        weight_cols += ["weight_nominal"]
    if systematic_weights:
        weight_cols += [c for c in df.columns if "weight_sys" in c]

    res = fix1dmw(df[var], df[weight_cols], bins=bins, range=range, flow=True, omp=True)
    res0 = pd.DataFrame(res[0], columns=weight_cols)
    res1 = pd.DataFrame(res[1], columns=weight_cols)
    res0._var_used = var
    res0._nbins = bins
    res0._xmin = range[0]
    res0._xmax = range[1]
    res1.var_used = var
    return (res0, res1)



"""
Command line application
"""

# stdlib
import argparse
import json
import pathlib
import logging

# external
import yaml

# tdub
from tdub.apply import FoldedResult, generate_npy
from tdub.batch import gen_submit_script
from tdub.features import create_parquet_files, prepare_from_parquet, FeatureSelector
from tdub.frames import raw_dataframe, drop_cols
from tdub.rex_art import run_stacks, run_pulls
from tdub.train import gp_minimize_auc, folded_training, prepare_from_root
from tdub.utils import SampleInfo, quick_files, override_features
from tdub import setup_logging


def parse_args():
    # fmt: off
    parser = argparse.ArgumentParser(prog="tdub", description="tdub CLI")
    subparsers = parser.add_subparsers(dest="action", help="")
    subparsers.metavar = "action           "
    formatter = lambda prog: argparse.HelpFormatter(prog, max_help_position=60)
    common_parser = argparse.ArgumentParser(add_help=False, formatter_class=formatter)
    common_parser.add_argument("--debug", action="store_true", help="set logging level to debug")

    applygennpy = subparsers.add_parser("apply-gennpy", help="Calculate samples BDT response and save to .npy file", parents=[common_parser])
    applygennpy.add_argument("--bnl-dir", type=str, help="directory on BNL parse for generating a condor submission script")
    applygennpy.add_argument("--single-file", type=str, help="input ROOT file")
    applygennpy.add_argument("--all-in-dir", type=str, help="Process all files in a directory")
    applygennpy.add_argument("-f", "--folds", type=str, nargs="+", help="directories with outputs from folded trainings", required=True)
    applygennpy.add_argument("-n", "--arr-name", type=str, default="pbdt_response", help="name for the array")
    applygennpy.add_argument("-o", "--out-dir", type=str, help="save output to a specific directory")
    applygennpy.add_argument("--bnl-script-name", type=str, help="name for output condor submission script when using '--bnl-dir")
    applygennpy.add_argument("--ignore-main", action="store_true", help="skip 'main' samples (ttbar & tW_(DR,DS) nominal FS)")

    rexpulls = subparsers.add_parser("rex-pulls", help="create matplotlib pull plots from TRExFitter output", parents=[common_parser])
    rexpulls.add_argument("workspace", type=str, help="TRExFitter workspace")
    rexpulls.add_argument("config", type=str, help="TRExFitter config")
    rexpulls.add_argument("-o", "--out-dir", type=str, help="output directory")
    rexpulls.add_argument("--no-text", action="store_true", help="don't print values on plots")

    rexstacks = subparsers.add_parser("rex-stacks", help="create matplotlib stack plots from TRExFitter output", parents=[common_parser])
    rexstacks.add_argument("workspace", type=str, help="TRExFitter workspace")
    rexstacks.add_argument("-o", "--out-dir", type=str, help="output directory for plots")
    rexstacks.add_argument("--lumi", type=str, default="139", help="Integrated lumi. for text")
    rexstacks.add_argument("--do-postfit", action="store_true", help="produce post fit plots as well")
    rexstacks.add_argument("--skip-regions", type=str, default=None, help="skip regions based on regex")
    rexstacks.add_argument("--band-style", type=str, choices=["hatch", "shade"], default="hatch", help="systematic band style")
    rexstacks.add_argument("--legend-ncol", type=int, choices=[1, 2], default=1, help="number of legend columns")

    trainfold = subparsers.add_parser("train-fold", help="Perform a folded training", parents=[common_parser])
    trainfold.add_argument("optimdir", type=str, help="directory containing optimization information")
    trainfold.add_argument("datadir", type=str, help="Directory with ROOT files")
    trainfold.add_argument("-o", "--out-dir", type=str, default="_folded", help="output directory for saving optimizatin results")
    trainfold.add_argument("-s", "--seed", type=int, default=414, help="random seed for folding")
    trainfold.add_argument("-n", "--n-splits", type=int, default=3, help="number of splits for folding")
    trainfold.add_argument("-r", "--esr", type=int, default=15, help="early stopping rounds")
    trainfold.add_argument("--override-features", type=str, help="a YAML file containing feature lists for overriding default features")

    trainoptimize = subparsers.add_parser("train-optimize", help="Gaussian processes minimization for HP optimization", parents=[common_parser])
    trainoptimize.add_argument("region", type=str, help="Region to train", choices=["1j1b", "2j1b", "2j2b"])
    trainoptimize.add_argument("nlomethod", type=str, help="NLO method samples to use", choices=["DR", "DS", "Both"])
    trainoptimize.add_argument("datadir", type=str, help="Directory with ROOT files")
    trainoptimize.add_argument("-o", "--out-dir", type=str, default="_optim", help="output directory for saving optimizatin results")
    trainoptimize.add_argument("-n", "--n-calls", type=int, default=15, help="number of calls for the optimization procedure")
    trainoptimize.add_argument("-r", "--esr", type=int, default=15, help="early stopping rounds for the training")
    trainoptimize.add_argument("--override-features", type=str, help="a YAML file containing feature lists for overriding default features")

    fselprepare = subparsers.add_parser("fsel-prepare", help="Prepare a set of parquet files for feature selection", parents=[common_parser])
    fselprepare.add_argument("-i", "--in", dest="indir", type=str, required=True, help="Directory containing ROOT files")
    fselprepare.add_argument("-o", "--out", dest="outdir", type=str, required=True, help="Output directory to save parquet files")
    fselprepare.add_argument("--entrysteps", type=str, required=False, help="entrysteps argument for create_parquet_files function")
    fselprepare.add_argument("--use-campaign-weight", action="store_true", help="propagate weight_campaign to weight_nominal")

    fselexecute = subparsers.add_parser("fsel-execute", help="Execute a round of feature selection", parents=[common_parser])
    fselexecute.add_argument("-i", "--in-pqdir", type=str, help="Directory containg the parquet files")
    fselexecute.add_argument("-n", "--nlo-method", type=str, choices=["DR", "DS"], required=True, help="tW NLO sample")
    fselexecute.add_argument("-r", "--region", type=str, choices=["1j1b", "2j1b", "2j2b"], required=True, help="Region to process")
    fselexecute.add_argument("-t", "--type", type=str, dest="itype", choices=["split", "gain"], required=True, help="importance type")
    fselexecute.add_argument("-o", "--out",  type=str, required=False, default="_auto", help="Output directory to save selection result (prepended with 'fsel_result.')")
    fselexecute.add_argument("--corr-threshold", dest="corrt", type=float, default=0.85, help="Correlation threshold")
    fselexecute.add_argument("--importance-n-fits", dest="nfits", type=int, default=5, help="Number of fitting rounds for importance calc.")
    fselexecute.add_argument("--max-features", dest="maxf", type=int, default=25, help="maximum number of features to test iteratively")
    fselexecute.add_argument("--exclude", type=str, nargs="+", help="features to exclude from consideration")
    fselexecute.add_argument("--weight-mean", type=float, default=1.0, help="scale weights such that the mean is this value")
    fselexecute.add_argument("--test-case", type=int, required=False, help="when testing, get a subset of each sample of this size")
    fselexecute.add_argument("--ttbar-frac", type=float, required=False, help="use a fraction of the ttbar sample (can't be used with --test-case")

    # fmt: on
    return (parser.parse_args(), parser)


def _optimize(args):
    if args.override_features:
        with pathlib.PosixPath(args.override_features).open("r") as f:
            override_features(yaml.safe_load(f))
    return gp_minimize_auc(
        args.datadir,
        args.region,
        args.nlomethod,
        output_dir=args.out_dir,
        n_calls=args.n_calls,
        esr=args.esr,
    )


def _fselprepare(args):
    create_parquet_files(args.indir, args.outdir, args.entrysteps, args.use_campaign_weight)


def _fselexecute(args):
    if args.out == "_auto":
        name = f"{args.nlo_method}_{args.region}_{args.itype}"
    else:
        name = args.out

    full_df, full_labels, full_weights = prepare_from_parquet(
        args.in_pqdir,
        region=args.region,
        nlo_method=args.nlo_method,
        weight_mean=args.weight_mean,
        test_case_size=args.test_case,
        ttbar_frac=args.ttbar_frac,
    )
    if args.exclude:
        drop_cols(full_df, *args.exclude)

    fs = FeatureSelector(
        df=full_df,
        labels=full_labels,
        weights=full_weights,
        importance_type=args.itype,
        name=name,
    )
    # fs.check_for_uniques()
    fs.check_importances(n_fits=args.nfits)
    fs.check_collinearity(threshold=args.corrt)
    fs.check_candidates(n=args.maxf)
    fs.check_iterative_remove_aucs(max_features=args.maxf)
    fs.check_iterative_add_aucs(max_features=args.maxf)
    fs.save_result()


def _foldedtraining(args):
    if args.override_features:
        with pathlib.PosixPath(args.override_features).open("r") as f:
            override_features(yaml.safe_load(f))
    with open(f"{args.optimdir}/summary.json", "r") as f:
        summary = json.load(f)
    nlo_method = summary["nlo_method"]
    qfiles = quick_files(f"{args.datadir}")
    if nlo_method == "DR":
        tW_files = qfiles["tW_DR"]
    elif nlo_method == "DS":
        tW_files = qfiles["tW_DS"]
    elif nlo_method == "Both":
        tW_files = qfiles["tW_DR"] + qfiles["tW_DS"]
        tW_files.sort()
    else:
        raise ValueError("nlo_method must be 'DR' or 'DS' or 'Both'")

    df, labels, weights = prepare_from_root(tW_files, qfiles["ttbar"], summary["region"])
    folded_training(
        df,
        labels,
        weights,
        summary["best_params"],
        {"verbose": 20, "early_stopping_rounds": args.esr},
        args.out_dir,
        summary["region"],
        kfold_kw={"n_splits": args.n_splits, "shuffle": True, "random_state": args.seed},
    )
    return 0


def _pred2npy(args):
    n_opts = 0
    if args.single_file is not None:
        n_opts += 1
    if args.all_in_dir is not None:
        n_opts += 1
    if args.bnl_dir is not None:
        n_opts += 1
    if n_opts != 1:
        raise ValueError(
            "must choose one (and only one) of '--single-file', '--all-in-dir', '--bnl-dir'"
        )

    if args.out_dir:
        outdir = pathlib.PosixPath(args.out_dir)
    else:
        outdir = pathlib.PosixPath(".")

    if args.bnl_dir is not None:
        gen_submit_script(
            args.bnl_dir, args.folds, outdir, args.arr_name, args.bnl_script_name
        )
        return 0

    frs = [FoldedResult(p) for p in args.folds]
    necessary_branches = ["OS", "elmu", "reg1j1b", "reg2j1b", "reg2j2b"]
    for fold in frs:
        necessary_branches += fold.features
    necessary_branches = sorted(set(necessary_branches), key=str.lower)
    log = logging.getLogger("tdub.apply")
    log.info("Loading necessary branches:")
    for nb in necessary_branches:
        log.info(f" - {nb}")

    def process_sample(sample_name):
        stem = pathlib.PosixPath(sample_name).stem
        sampinfo = SampleInfo(stem)
        tree = f"WtLoop_{sampinfo.tree}"
        df = raw_dataframe(sample_name, tree=tree, branches=necessary_branches)
        npyfilename = outdir / f"{stem}.{args.arr_name}.npy"
        generate_npy(frs, df, npyfilename)

    main_samples_pfxs = (
        "tW_DR_410648_FS",
        "tW_DR_410649_FS",
        "tW_DS_410656_FS",
        "tW_DS_410657_FS",
        "ttbar_410472_FS",
    )

    if args.single_file is not None:
        process_sample(args.single_file)
        return 0
    elif args.all_in_dir is not None:
        for child in pathlib.PosixPath(args.all_in_dir).iterdir():
            if args.ignore_main and child.name.startswith(main_samples_pfxs):
                continue
            if child.suffix == ".root":
                process_sample(str(child.resolve()))
        return 0


def cli():
    args, parser = parse_args()
    if args.action is None:
        parser.print_help()
        return 0

    setup_logging()

    for name in logging.root.manager.loggerDict:
        if args.debug:
            logging.getLogger(name).setLevel(logging.DEBUG)
        else:
            logging.getLogger(name).setLevel(logging.INFO)

    # fmt: off
    if args.action == "rex-stacks":
        return run_stacks(args)
    elif args.action == "rex-pulls":
        return run_pulls(args)
    elif args.action == "train-optimize":
        return _optimize(args)
    elif args.action == "train-fold":
        return _foldedtraining(args)
    elif args.action == "apply-gennpy":
        return _pred2npy(args)
    elif args.action == "fsel-prepare":
        return _fselprepare(args)
    elif args.action == "fsel-execute":
        return _fselexecute(args)
    else:
        parser.print_help()
    # fmt: on



"""
tW analysis tools
"""

__version__ = "0.0.18"


def setup_logging():
    import logging

    # fmt: off
    logging.basicConfig(
        level=logging.INFO, format="{:22}  %(levelname)8s  %(message)s".format("[%(name)s : %(funcName)s]")
    )
    logging.addLevelName(logging.WARNING, "\033[1;31m{:8}\033[1;0m".format(logging.getLevelName(logging.WARNING)))
    logging.addLevelName(logging.ERROR, "\033[1;35m{:8}\033[1;0m".format(logging.getLevelName(logging.ERROR)))
    logging.addLevelName(logging.INFO, "\033[1;32m{:8}\033[1;0m".format(logging.getLevelName(logging.INFO)))
    logging.addLevelName(logging.DEBUG, "\033[1;34m{:8}\033[1;0m".format(logging.getLevelName(logging.DEBUG)))
    # fmt: on



"""
Module for applying trained models
"""

# stdlib
import json
import logging
import os
from pathlib import PosixPath
from typing import List, Dict, Any, Union

# external
import numpy as np
import joblib
import pandas as pd
from sklearn.model_selection import KFold

try:
    import lightgbm as lgbm
except ImportError:
    class lgbm:
        LGBMClassifier = None

# tdub
from tdub.frames import specific_dataframe
from tdub.utils import Region, get_selection


log = logging.getLogger(__name__)


class FoldedResult:
    """Provides access to the properties of a folded training result

    Parameters
    ----------
    fold_output : str
       the directory with the folded training output

    Attributes
    ----------
    model0 : lightgbm.LGBMClassifier
       the model for the 0th fold from training
    model1 : lightgbm.LGBMClassifier
       the model for the 1st fold from training
    model2 : lightgbm.LGBMClassifier
       the model for the 2nd fold from training
    region : Region
       the region for this training
    features : list(str)
       the list of kinematic features used by the model
    folder : sklearn.model_selection.KFold
       the folding object that the training session used
    summary : dict(str, Any)
       the contents of the ``summary.json`` file.

    Examples
    --------

    >>> from tdub.apply import FoldedResult
    >>> fr_1j1b = FoldedResult("/path/to/folded_training_1j1b")

    """

    def __init__(self, fold_output: str) -> None:
        fold_path = PosixPath(fold_output)
        if not fold_path.exists():
            raise ValueError(f"{fold_output} does not exist")
        fold_path = fold_path.resolve()
        self._model0 = joblib.load(fold_path / "model_fold0.joblib.gz")
        self._model1 = joblib.load(fold_path / "model_fold1.joblib.gz")
        self._model2 = joblib.load(fold_path / "model_fold2.joblib.gz")

        summary_file = fold_path / "summary.json"
        self._summary = json.loads(summary_file.read_text())
        self._features = self._summary["features"]
        self._folder = KFold(**(self._summary["kfold"]))
        self._region = Region.from_str(self._summary["region"])

    @property
    def model0(self) -> lgbm.LGBMClassifier:
        return self._model0

    @property
    def model1(self) -> lgbm.LGBMClassifier:
        return self._model1

    @property
    def model2(self) -> lgbm.LGBMClassifier:
        return self._model2

    @property
    def features(self) -> List[str]:
        return self._features

    @property
    def region(self) -> Region:
        return self._region

    @property
    def folder(self) -> KFold:
        return self._folder

    @property
    def summary(self) -> Dict[str, Any]:
        return self._summary

    def to_files(
        self, files: Union[str, List[str]], tree: str = "WtLoop_nominal"
    ) -> np.ndarray:
        """apply the folded result to a set of files

        Parameters
        ----------
        files : str or list(str)
          the input file(s) to open and apply to
        tree : str
          the name of the tree to extract data from

        Returns
        -------
        numpy.ndarray
          the classifier output for the region associated with ``fr``

        Examples
        --------
        >>> from tdub.apply import FoldedResult
        >>> fr_1j1b = FoldedResult("/path/to/folded_training_1j1b")
        >>> y = fr_1j1b.to_files(["/path/to/file1.root", "/path/to/file2.root"])

        """
        dfim = specific_dataframe(files, self.region, tree=tree, to_ram=True)
        dfim._df = dfim.df[self.features]

        X = dfim.df.to_numpy()
        y0 = self.model0.predict_proba(X)[:, 1]
        y1 = self.model1.predict_proba(X)[:, 1]
        y2 = self.model2.predict_proba(X)[:, 1]
        y = np.mean([y0, y1, y2], axis=0)

        return y

    def to_dataframe(
        self,
        df: pd.DataFrame,
        column_name: str = "unnamed_bdt_response",
        query: bool = False,
    ) -> None:
        """apply trained models to an arbitrary dataframe.

        This function will augment the dataframe with a new column
        (with a name given by the ``column_name`` argument) if it
        doesn't already exist. If the dataframe is empty this function
        does nothing.

        Parameters
        ----------
        df : pandas.DataFrame
           the dataframe to read and augment
        column_name : str
           name to give the BDT response variable
        query : bool
           perform a query on the dataframe to select events belonging to
           the region associated with ``fr`` necessary if the dataframe
           hasn't been pre-filtered

        Examples
        --------

        >>> from tdub.apply import FoldedResult
        >>> from tdub.frames import conservative_dataframe
        >>> df = conservative_dataframe("/path/to/file.root")
        >>> fr_1j1b = FoldedResult("/path/to/folded_training_1j1b")
        >>> fr_1j1b.to_dataframe(df, query=True)

        """
        if df.shape[0] == 0:
            log.info("Dataframe is empty, doing nothing")
            return None

        if column_name not in df.columns:
            log.info(f"Creating {column_name} column")
            df[column_name] = -9999.0

        if query:
            log.info(f"applying selection filter [ {get_selection(self.region)} ]")
            mask = df.eval(get_selection(self.region))
            X = df[self.features].to_numpy()[mask]
        else:
            X = df[self.features].to_numpy()

        if X.shape[0] == 0:
            return None

        y0 = self.model0.predict_proba(X)[:, 1]
        y1 = self.model1.predict_proba(X)[:, 1]
        y2 = self.model2.predict_proba(X)[:, 1]
        y = np.mean([y0, y1, y2], axis=0)

        if query:
            df.loc[mask, column_name] = y
        else:
            df[column_name] = y


def generate_npy(
    frs: List[FoldedResult], df: pd.DataFrame, output_file: Union[str, os.PathLike]
) -> None:
    """create a NumPy npy file which is the response for all events in a DataFrame

    this will use all folds in the ``frs`` argument to get BDT
    response any each region associated to a ``FoldedResult``. We
    query the input df to ensure that we apply to the correct
    event. If the input dataframe is empty (no rows) then an empty
    array is written to disk.

    Parameters
    ----------
    frs : list(FoldedResult)
       the folded results to use
    df : pandas.DataFrame
       the dataframe of events to get the responses for
    output_file : str or os.PathLike
       name of the output NumPy file

    Examples
    --------

    >>> from tdub.apply import FoldedResult, generate_npy
    >>> from tdub.frames import conservative_dataframe
    >>> df = conservative_dataframe("/path/to/file.root")
    >>> fr_1j1b = FoldedResult("/path/to/folded_training_1j1b")
    >>> fr_2j1b = FoldedResult("/path/to/folded_training_2j1b")
    >>> fr_2j2b = FoldedResult("/path/to/folded_training_2j2b")
    >>> generate_npy([fr_1j1b, fr_2j1b, fr_2j2b], df, "output.npy")

    """

    if df.shape[0] == 0:
        log.info(f"Saving empty array to {output_file}")
        np.save(output_file, np.array([], dtype=np.float64))
        return None

    outfile = PosixPath(output_file)
    outfile.parent.mkdir(parents=True, exist_ok=True)

    colname = "_temp_col"
    log.info(f"The {colname} column will be deleted at the end of this function")
    for fr in frs:
        fr.to_dataframe(df, column_name=colname, query=True)
    np.save(outfile, df[colname].to_numpy())
    log.info(f"Saved output to {outfile}")
    df.drop(columns=[colname], inplace=True)
    log.info(f"Temporary column '{colname}' deleted")
